#
#  SPDX-FileCopyrightText: 2012 Sebastian Gottfried <sebastiangottfried@web.de>
#
#  SPDX-License-Identifier: GPL-2.0-or-later
#

import argparse
import os
import uuid
from lxml import etree

class Course(object):
    def __init__(self, id, title, description, keyboard_layout_name, lessons=None):
        self.id = id
        self.title = title
        self.description = description
        self.keyboard_layout_name = keyboard_layout_name
        self.lessons = lessons or []
    def xml_tree(self):
        root = etree.Element('course')
        id = etree.SubElement(root, "id")
        id.text = self.id
        title = etree.SubElement(root, "title")
        title.text = self.title
        description = etree.SubElement(root, "description")
        description.text = self.description
        keyboard_layout_name = etree.SubElement(root, "keyboardLayout")
        keyboard_layout_name.text = self.keyboard_layout_name
        lessons = etree.SubElement(root, "lessons")
        for lesson in self.lessons:
            lesson.build_xml(lessons)
        return root
    def __repr__(self):
        return "<Course {!r}, {!r}, {!r}, {!r}>".format(
            self.title,
            self.description,
            self.keyboard_layout_name,
            self.lessons
        )

class Lesson(object):
    def __init__(self, id, title, new_characters, text):
        self.id = id
        self.title = title
        self.new_characters = new_characters
        self.text = text
    def build_xml(self, parent_node):
        node = etree.SubElement(parent_node, "lesson")
        id = etree.SubElement(node, "id")
        id.text = self.id
        title = etree.SubElement(node, "title")
        title.text = self.title
        new_characters = etree.SubElement(node, "newCharacters")
        new_characters.text = self.new_characters
        text = etree.SubElement(node, "text")
        text.text = self.text
    def __repr__(self):
        return "<Lesson {!r}, {!r}, {!r}>".format(
            self.title,
            self.new_characters,
            self.text
        )

def read(f, layout):
    tree = etree.parse(f)
    root = tree.getroot()
    id = make_id()
    title, = root.xpath("//KTouchLecture/Title[1]/text()")
    description, = root.xpath("//KTouchLecture/Comment[1]/text()")
    lesson_nodes = root.xpath("//KTouchLecture/Levels/Level")
    lessons = [parse_lesson(node) for node in lesson_nodes]
    return Course(id, title, description, layout, lessons)

def parse_lesson(lesson_node):
    id = make_id()
    title, = lesson_node.xpath("./NewCharsLabel[1]/text()") or ('',)
    if title == '':
        index = lesson_node.getparent().index(lesson_node)
        title = u"Lesson {}".format(index + 1)
    new_characters, = lesson_node.xpath("./NewCharacters[1]/text()") or ('',)
    input_lines = lesson_node.xpath("./Line/text()")
    lines = []
    for line in input_lines:
        line = unicode(line)
        while len(line) > 59:
            try:
                split_pos = line.rindex(" ", 0, 60)
            except ValueError:
                break
            lines.append(line[:split_pos])
            line = line[split_pos + 1:]
        lines.append(line)
    text = u"\n".join(lines)
    return Lesson(id, title, new_characters, text)

def parse_chars(char_nodes):
    chars = []
    for char_node in char_nodes:
        value = char_node.text
        position = POSITION_MAP[char_node.attrib['Position']]
        modifier = MODIFIER_MAP[position]
        char = Char(value, position, modifier)
        chars.append(char)
        if value.lower() != value:
            hiddenChar = Char(value.lower(), 'hidden')
            char.modifier = 'shift'
            chars.append(hiddenChar)
    return chars


def make_id():
    return "{{{}}}".format(uuid.uuid4())

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='convert ktouch course files'
    )
    parser.add_argument('-l', '--layout',
        type=str,
        metavar="LAYOUT",
        default=""
    )
    parser.add_argument('input_file',
        type=argparse.FileType('r'),
        metavar="INPUTFILE",
    )
    args = parser.parse_args()
    course = read(args.input_file, args.layout)
    print etree.tostring(course.xml_tree())



#
#  Copyright 2012  Sebastian Gottfried <sebastiangottfried@web.de>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import argparse, os
from lxml import etree

def find_courses(data_dir, rel_path, root):
    path = os.path.join(data_dir, rel_path)
    course_elems = []
    for filename in os.listdir(path):
        if filename.endswith('.xml'):
            course_tree = etree.parse(file(os.path.join(path, filename), 'r'))
            elem = etree.Element("course")
            title = etree.SubElement(elem, "title")
            title.text, = course_tree.xpath("//course/title/text()")
            description = etree.SubElement(elem, "description")
            description.text, = course_tree.xpath("//course/description/text()")
            keyboard_layout_name = etree.SubElement(elem, "keyboardLayout")
            keyboard_layout_name.text, = course_tree.xpath("//course/keyboardLayout/text()")
            id_ele = etree.SubElement(elem, "id")
            id_ele.text, = course_tree.xpath("//course/id/text()")
            path_ele = etree.SubElement(elem, "path")
            path_ele.text = os.path.join(rel_path, filename)
            course_elems.append(elem)
    course_elems.sort(key=lambda elem: elem.xpath("path/text()"))
    for elem in course_elems:
        root.append(elem)

def find_keyboard_layouts(data_dir, rel_path, root):
    path = os.path.join(data_dir, rel_path)
    keyboard_layout_elems = []
    for filename in os.listdir(path):
        if filename.endswith('.xml'):
            keyboard_tree = etree.parse(file(os.path.join(path, filename), 'r'))
            elem = etree.Element("keyboardLayout")
            title = etree.SubElement(elem, "title")
            title.text, = keyboard_tree.xpath("//keyboardLayout/title/text()")
            name = etree.SubElement(elem, "name")
            name.text, = keyboard_tree.xpath("//keyboardLayout/name/text()")
            id_ele = etree.SubElement(elem, "id")
            id_ele.text, = keyboard_tree.xpath("//keyboardLayout/id/text()")
            path_ele = etree.SubElement(elem, "path")
            path_ele.text = os.path.join(rel_path, filename)
            keyboard_layout_elems.append(elem)
    keyboard_layout_elems.sort(key=lambda elem: elem.xpath("path/text()"))
    for elem in keyboard_layout_elems:
        root.append(elem)
            

if __name__ == '__main__':
    def dir_arg(path):
        if not os.path.exists(path):
            msg = "{!r} doesn't exist".format(path)
            raise argparse.ArgumentTypeError(msg)
        if not os.path.isdir(path):
            msg = "{!r} has to be a directory".format(path)
            raise argparse.ArgumentTypeError(msg)
        return path
    parser = argparse.ArgumentParser(
        description='builds data.xml index file'
    )
    parser.add_argument("datadir",
        type=dir_arg,
        help="the directory containing the data files",
        metavar="DIR"
    )
    args = parser.parse_args()
    root = etree.Element('data')
    find_courses(args.datadir, 'courses', root)
    find_keyboard_layouts(args.datadir, 'keyboardlayouts', root)
    print etree.tostring(root)



#
#  SPDX-FileCopyrightText: 2012 Sebastian Gottfried <sebastiangottfried@web.de>
#
#  SPDX-License-Identifier: GPL-2.0-or-later
#

import argparse
import uuid
from lxml import etree

class KeyboardLayout(object):
    def __init__(self, id, title, name, keys=None):
        self.id = id
        self.title = title
        self.name = name
        self.keys = keys or []
    @property
    def keys(self):
        return self._keys
    @keys.setter
    def keys(self, keys):
        self._keys = keys
        self.width = 0
        self.height = 0
        min_left = None
        min_top = None
        for key in keys:
            if key.left < min_left or min_left is None:
                min_left = key.left
            if key.top < min_top or min_top is None:
                min_top = key.top
        for key in keys:
            key.left = key.left - min_left
            key.top = key.top - min_top
            if key.left + key.width > self.width:
                self.width = key.left + key.width
            if key.top + key.height > self.height:
                self.height = key.top + key.height
    def xml_tree(self):
        root = etree.Element('keyboardLayout')
        id = etree.SubElement(root, "id")
        id.text = self.id
        title = etree.SubElement(root, "title")
        title.text = self.title
        name = etree.SubElement(root, "name")
        name.text = self.name
        width = etree.SubElement(root, "width")
        width.text = str(self.width)
        height = etree.SubElement(root, "height")
        height.text = str(self.height)
        keys = etree.SubElement(root, "keys")
        for key in self.keys:
            if isinstance(key, Key):
                key.build_xml(keys)
        for key in self.keys:
            if isinstance(key, SpecialKey):
                key.build_xml(keys)
        return root
    def __repr__(self):
        return "<KeyboardLayout {!r}, {!r}, {!r}, {!r}, {!r}>".format(
            self.title,
            self.name,
            self.keys,
            self.width,
            self.height
        )

class AbstractKey(object):
    def __init__(self, left, top, width, height):
        self.left = left
        self.top = top
        self.width = width
        self.height = height
        
class Key(AbstractKey):
    def __init__(self, left, top, width, height, finger_index, chars=None):
        super(Key, self).__init__(left, top, width, height)
        self.finger_index = finger_index
        self.chars = chars or []
        self.has_haptic_marker = False
    def build_xml(self, parent_node):
        node = etree.SubElement(parent_node, "key")
        node.attrib['left'] = str(self.left)
        node.attrib['top'] = str(self.top)
        node.attrib['width'] = str(self.width)
        node.attrib['height'] = str(self.height)
        node.attrib['fingerIndex'] = str(self.finger_index)
        if self.has_haptic_marker:
            node.attrib['hasHapticMarker'] = "true"
        for char in self.chars:
            char.build_xml(node)
    def __repr__(self):
        return "<Key {!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r}>".format(
            self.left,
            self.top,
            self.width,
            self.height,
            self.finger_index,
            self.chars,
            self.has_haptic_marker
        )

class SpecialKey(AbstractKey):
    def __init__(self, left, top, width, height, type):
        super(SpecialKey, self).__init__(left, top, width, height)
        self.type = type
        self.modifier_id = None
        self.label = None
    def build_xml(self, parent_node):
        node = etree.SubElement(parent_node, "specialKey")
        node.attrib['left'] = str(self.left)
        node.attrib['top'] = str(self.top)
        node.attrib['width'] = str(self.width)
        node.attrib['height'] = str(self.height)
        node.attrib['type'] = str(self.type)
        if self.modifier_id is not None:
            node.attrib['modifierId'] = self.modifier_id
        if self.label is not None:
            node.attrib['label'] = self.label
    def __repr__(self):
        return "<SpecialKey {!r}, {!r}, {!r}, {!r}, {!r}, {!r}, {!r}>".format(
            self.left,
            self.top,
            self.width,
            self.height,
            self.type,
            self.modifier_id,
            self.label
        )

class Char(object):
    def __init__(self, value, position, modifier=None):
        self.value = value
        self.position = position
        self.modifier = modifier
    def build_xml(self, parent_node):
        node = etree.SubElement(parent_node, "char")
        node.text = self.value
        node.attrib['position'] = self.position
        if self.modifier is not None:
            node.attrib['modifier'] = str(self.modifier)
    def __repr__(self):
        return u"<Char {!r}, {!r}, {!r}>".format(
            self.value,
            self.position,
            self.modifier
        )

POSITION_MAP = {
    "TopLeft": "topLeft",
    "TopRight": "topRight",
    "BottomLeft": "bottomLeft",
    "BottomRight": "bottomRight"
}

MODIFIER_MAP = {
    "topLeft": "shift",
    "topRight": None,
    "bottomLeft": None,
    "bottomRight": "altgr"
}

def read(f):
    tree = etree.parse(f)
    root = tree.getroot()
    id = make_id()
    title, = etree.XPath("//KTouchKeyboard/Title[1]/text()")(root)
    name, = etree.XPath("//KTouchKeyboard/Language[1]/text()")(root)
    key_nodes = etree.XPath("//KTouchKeyboard/Keys/Key")(root)
    keys = [parse_key(node) for node in key_nodes]
    return KeyboardLayout(id, title, name, keys)
    
def parse_key(key_node):
    get_char_nodes = etree.XPath("./Char")
    type = key_node.attrib['Type']
    left = int(key_node.attrib['X'])
    top = int(key_node.attrib['Y'])
    width = int(key_node.attrib['Width'])
    height = int(key_node.attrib['Height'])
    if type in ['NORMAL', 'FINGER']:
        has_haptic_marker = False
        if type == 'FINGER':
            finger_index = key_node.getparent().index(key_node)
            if finger_index in [3, 4]:
                has_haptic_marker = True
        else:
            finger_index = key_node.attrib['FingerKeyIndex']
        
        chars = parse_chars(get_char_nodes(key_node), finger_index)
        key = Key(left, top, width, height, finger_index, chars)
        key.has_haptic_marker = has_haptic_marker
    elif type == "TAB":
        key = SpecialKey(left, top, width, height, 'tab')
    elif type == "CAPSLOCK":
        key = SpecialKey(left, top, width, height, 'capslock')
    elif type == "SHIFT":
        key = SpecialKey(left, top, width, height, 'shift')
        key.modifier_id = left < 700 and 'left_shift' or 'right_shift'
    elif type == "BACKSPACE":
        key = SpecialKey(left, top, width, height, 'backspace')
    elif type == "ENTER":
        key = SpecialKey(left, top, width, height, 'return')
    elif type == "SPACE":
        key = SpecialKey(left, top, width, height, 'space')
    else:
        key = SpecialKey(left, top, width, height, 'other')
        key.label = key_node.attrib['KeyText']
        if key.label == "AltGr":
            key.label = "Alt Gr"
        if key.label == "Alt Gr":
            key.modifier_id = 'altgr'
    return key

def parse_chars(char_nodes, finger_index):
    chars = []
    for char_node in char_nodes:
        value = char_node.text
        position = POSITION_MAP[char_node.attrib['Position']]
        modifier = MODIFIER_MAP[position]
        if modifier == 'shift':
            modifier = finger_index <= 4 and "right_shift" or "left_shift"
        char = Char(value, position, modifier)
        chars.append(char)
        if value.lower() != value:
            hiddenChar = Char(value.lower(), 'hidden')
            char.modifier = finger_index <= 4 and "right_shift" or "left_shift"
            chars.append(hiddenChar)
    return chars

def make_id():
    return "{{{}}}".format(uuid.uuid4())

if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        description='convert ktouch keyboard files'
    )
    parser.add_argument('input_file',
        type=argparse.FileType('r'),
        metavar="INPUTFILE"
    )
    args = parser.parse_args()
    layout = read(args.input_file)
    print etree.tostring(layout.xml_tree())    
